from typing import Union
import pypiper
import os
from pathlib import Path

import logging
import tempfile
import gzip
import shutil

from refgenconf.exceptions import MissingGenomeError

from ubiquerg import is_command_callable
from geniml.io import RegionSet

from geniml.bbclient import BBClient

from bedboss.bedclassifier import get_bed_type
from bedboss.bedqc.bedqc import bedqc
from bedboss.exceptions import RequirementsException, BedBossException

from bedboss.bedmaker.const import (
    BEDGRAPH_TEMPLATE,
    BIGWIG_TEMPLATE,
    BIGBED_TEMPLATE,
    WIG_TEMPLATE,
    BED_TO_BIGBED_PROGRAM,
    BIGBED_TO_BED_PROGRAM,
    QC_FOLDER_NAME,
    BIGBED_FILE_NAME,
)
from bedboss.bedmaker.utils import get_chrom_sizes
from bedboss.bedmaker.models import InputTypes, BedMakerOutput

_LOGGER = logging.getLogger("bedboss")


def make_bigbed(
    bed_path: Union[str, Path],
    output_path: Union[str, Path],
    genome: str,
    bed_type: str = None,
    rfg_config: Union[str, Path] = None,
    chrom_sizes: Union[str, Path] = None,
    pm: pypiper.PipelineManager = None,
) -> str:
    """
    Generate bigBed file for the BED file.

    :param bed_path: path to the BED file
    :param output_path: parent folder to save the bigBed file. (function will create a subfolder for bigBed files)
    :param genome: reference genome (e.g. hg38, mm10, etc.)
    :param chrom_sizes: a full path to the chrom.sizes required for the
                        bedtobigbed conversion
    :param rfg_config: file path to the genome config file. [Default: None]
    :param bed_type: bed type to be used for bigBed file generation "bed{bedtype}+{n}" [Default: None] (e.g bed3+1)
    :param pm: pypiper object

    :return: path to the bigBed file
    """
    if not pm:
        pm = pypiper.PipelineManager(
            name="bedmaker",
            outfolder=os.path.join(output_path, "bedmaker_logs"),
            recover=True,
            multi=True,
        )
        pm_clean = True
    else:
        pm_clean = False
    _LOGGER.info(f"Generating bigBed files for: {bed_path}")

    bigbed_output_folder = os.path.join(output_path, BIGBED_FILE_NAME)
    if not os.path.exists(bigbed_output_folder):
        os.makedirs(bigbed_output_folder)

    bedfile_name = os.path.split(bed_path)[1]
    fileid = os.path.splitext(os.path.splitext(bedfile_name)[0])[0]
    # Produce bigBed (big_narrow_peak) file from peak file
    big_bed_path = os.path.join(bigbed_output_folder, fileid + ".bigBed")
    if not chrom_sizes:
        try:
            chrom_sizes = get_chrom_sizes(genome=genome, rfg_config=rfg_config)
        except MissingGenomeError:
            _LOGGER.error(f"Could not find Genome in refgenie. Skipping...")
            chrom_sizes = ""

    temp = os.path.join(output_path, next(tempfile._get_candidate_names()))

    if not os.path.exists(big_bed_path):
        pm.clean_add(temp)

        if not is_command_callable(f"{BED_TO_BIGBED_PROGRAM}"):
            raise RequirementsException(
                "To convert bed to BigBed file You must first install "
                "bedToBigBed add in your PATH. "
                "Instruction: "
                "https://genome.ucsc.edu/goldenpath/help/bigBed.html"
            )
        if bed_type is not None:
            cmd = f"zcat {bed_path} | sort -k1,1 -k2,2n > {temp}"
            pm.run(cmd, temp, nofail=False)

            cmd = f"{BED_TO_BIGBED_PROGRAM} -type={bed_type} {temp} {chrom_sizes} {big_bed_path}"
            try:
                _LOGGER.info(f"Running: {cmd}")
                pm.run(cmd, big_bed_path, nofail=False)
            except Exception as err:
                raise BedBossException(
                    f"Fail to generating bigBed files for {bed_path}: "
                    f"unable to validate genome assembly with Refgenie. "
                    f"Error: {err}"
                )
        else:
            cmd = (
                "zcat "
                + bed_path
                + " | awk '{ print $1, $2, $3 }'| sort -k1,1 -k2,2n > "
                + temp
            )
            pm.run(cmd, temp)
            cmd = f"{BED_TO_BIGBED_PROGRAM} -type=bed3 {temp} {chrom_sizes} {big_bed_path}"

            try:
                pm.run(cmd, big_bed_path, nofail=True)
            except Exception as err:
                _LOGGER.info(
                    f"Fail to generating bigBed files for {bed_path}: "
                    f"unable to validate genome assembly with Refgenie. "
                    f"Error: {err}"
                )
        pm._cleanup()
    if pm_clean:
        pm.stop_pipeline()
    return big_bed_path


def make_bed(
    input_file: str,
    input_type: str,
    output_path: str,
    genome: str,
    narrowpeak: bool = False,
    rfg_config: str = None,
    chrom_sizes: str = None,
    pm: pypiper.PipelineManager = None,
) -> str:
    """
    Convert the input file to BED format by construct the command based
    on input file type and execute the command.

    :param input_file: path to the input file
    :param input_type: a [bigwig|bedgraph|bed|bigbed|wig] file that will be
                        converted into BED format
    :param output_path: path to the output folder, logs will be saved (BED files will be cached)
    :param genome: reference genome
    :param narrowpeak: whether the regions are narrow (transcription factor
                        implies narrow, histone mark implies broad peaks)
    :param rfg_config: file path to the genome config file
    :param chrom_sizes: a full path to the chrom.sizes required for the
                        bedtobigbed conversion
    :param pm: pypiper object

    :return: path to the BED file
    """
    _LOGGER.info(f"Converting {os.path.abspath(input_file)} to BED format.")

    input_type = input_type.lower()
    if input_type not in [member.value for member in InputTypes]:
        raise BedBossException(
            f"Invalid input type: {input_type}. "
            f"Supported types: {', '.join([k.value for k in InputTypes])}"
        )

    if not pm:
        pm = pypiper.PipelineManager(
            name="bedmaker",
            outfolder=os.path.join(os.path.dirname(output_path), "bedmaker_logs"),
            recover=True,
            multi=True,
        )
        pm_clean = True
    else:
        pm_clean = False

    temp_bed_path = os.path.splitext(output_path)[0]

    file_base_name = os.path.basename(input_file)
    input_extension = os.path.splitext(file_base_name)[1]

    width = "bdgbroadcall" if not narrowpeak else "bdgpeakcall"

    # creat cmd to run that convert non bed file to bed file
    if input_type == InputTypes.BED.value:
        try:
            # If bed file was provided:
            bbclient = BBClient()
            bed_id = bbclient.add_bed_to_cache(input_file)
            output_path = bbclient.seek(bed_id)
        except FileNotFoundError as e:
            raise BedBossException(f"File not found: {input_file} Error: {e}")

    else:
        _LOGGER.info(f"Converting {input_file} to BED format")

        # Use the gzip and shutil modules to produce temporary unzipped files
        if input_extension == ".gz":
            temp_input_file = os.path.join(
                os.path.dirname(output_path),
                os.path.splitext(file_base_name)[0],
            )
            with gzip.open(input_file, "rb") as f_in:
                with open(temp_input_file, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
            pm.clean_add(temp_input_file)

        # creating cmd for bedGraph files
        if input_type == InputTypes.BED_GRAPH.value:
            if not is_command_callable("macs2"):
                raise RequirementsException(
                    "To convert bedGraph file You must first install "
                    "macs2 and add it to your PATH. "
                    "Instruction: "
                    "https://pypi.org/project/MACS2/"
                )
            else:
                cmd = BEDGRAPH_TEMPLATE.format(
                    input=input_file,
                    output=temp_bed_path,
                    width=width,
                )

        # creating cmd for bigWig files
        elif input_type == InputTypes.BIG_WIG.value:
            if not is_command_callable("bigWigToBedGraph"):
                raise RequirementsException(
                    "To convert bigWig file You must first install "
                    "bigWigToBedGraph and add it to your PATH. "
                    "Instruction: "
                    "https://genome.ucsc.edu/goldenpath/help/bigWig.html"
                )
            else:
                cmd = BIGWIG_TEMPLATE.format(
                    input=input_file,
                    output=temp_bed_path,
                    width=width,
                )
        # creating cmd for wig files
        elif input_type == InputTypes.WIG.value:
            if not chrom_sizes:
                chrom_sizes = get_chrom_sizes(genome=genome, rfg_config=rfg_config)

            # define a target for temporary bw files
            temp_target = os.path.join(output_path, file_base_name + ".bw")
            if not is_command_callable("wigToBigWig"):
                raise RequirementsException(
                    "To convert wig file You must first install "
                    "wigToBigWig and add it in your PATH. "
                    "Instruction: "
                    "https://genome.ucsc.edu/goldenpath/help/bigWig.html"
                )
            else:
                pm.clean_add(temp_target)
                cmd1 = WIG_TEMPLATE.format(
                    input=input_file,
                    intermediate_bw=temp_target,
                    chrom_sizes=chrom_sizes,
                    width=width,
                )

            if not is_command_callable("bigWigToBedGraph"):
                raise RequirementsException(
                    "To convert bigWig file You must first install "
                    "bigWigToBedGraph and add it in your PATH. "
                    "Instruction: "
                    "https://genome.ucsc.edu/goldenpath/help/bigWig.html"
                )
            else:
                cmd2 = BIGWIG_TEMPLATE.format(
                    input=temp_target,
                    output=temp_bed_path,
                    width=width,
                )
                cmd = [cmd1, cmd2]
        # creating cmd for bigBed files
        elif input_type == InputTypes.BIG_BED.value:
            if not is_command_callable(BIGBED_TO_BED_PROGRAM):
                raise RequirementsException(
                    "To convert bigBed file You must first install "
                    "bigBedToBed and add it in your PATH. "
                    "Instruction: "
                    "https://genome.ucsc.edu/goldenpath/help/bigBed.html"
                )
            else:
                cmd = BIGBED_TEMPLATE.format(input=input_file, output=temp_bed_path)

        else:
            raise NotImplementedError(f"'{input_type}' format is not supported")

        pm.run(cmd, temp_bed_path, nofail=False)

        bbclient = BBClient()
        bed_id = bbclient.add_bed_to_cache(input_file)
        output_path = bbclient.seek(bed_id)

    pm._cleanup()
    if pm_clean:
        pm.stop_pipeline()

    return output_path


def make_all(
    input_file: str,
    input_type: str,
    output_path: str,
    genome: str,
    rfg_config: str = None,
    chrom_sizes: str = None,
    narrowpeak: bool = False,
    check_qc: bool = True,
    pm: pypiper.PipelineManager = None,
) -> BedMakerOutput:
    """
    Maker of bed and bigbed files.

    Pipeline to convert supported file formats into
    BED format and bigBed format. Currently supported formats*:
        - bedGraph
        - bigBed
        - bigWig
        - wig
        - bed
    :param input_file: path to the input file
    :param input_type: a [bigwig|bedgraph|bed|bigbed|wig] file that will be
                       converted into BED format
    :param output_path: path to the output folder, where bigbed and logs will be saved
    :param genome: reference genome
    :param rfg_config: file path to the genome config file
    :param chrom_sizes: a full path to the chrom.sizes required for the
                        bedtobigbed conversion
    :param narrowpeak: whether the regions are narrow (transcription factor
                       implies narrow, histone mark implies broad peaks)
    :param check_qc: run quality control during bedmaking
    :param pm: pypiper object

    :return: dict with generated bed metadata - BedMakerOutput object:
        {
            "bed_type": bed_type. e.g. bed, bigbed
            "bed_format": bed_format. e.g. narrowpeak, broadpeak
            "bed_file": path to the bed file
            "bigbed_file": path to the bigbed file
            "bed_digest": bed_digest
        }
    """
    if not pm:
        pm = pypiper.PipelineManager(
            name="bedmaker",
            outfolder=os.path.join(output_path, "bedmaker_logs"),
            recover=True,
            multi=True,
        )
        pm_clean = True
    else:
        pm_clean = False
    output_bed = make_bed(
        input_file=input_file,
        input_type=input_type,
        output_path=output_path,
        genome=genome,
        narrowpeak=narrowpeak,
        rfg_config=rfg_config,
        chrom_sizes=chrom_sizes,
        pm=None,
    )
    bed_type, bed_format = get_bed_type(output_bed)
    if check_qc:
        try:
            bedqc(
                output_bed,
                outfolder=os.path.join(output_path, QC_FOLDER_NAME),
                pm=pm,
            )
        except Exception as e:
            raise BedBossException(
                f"Quality control failed for {output_path}. Error: {e}"
            )
    try:
        output_bigbed = make_bigbed(
            bed_path=output_bed,
            output_path=output_path,
            genome=genome,
            bed_type=bed_type,
            rfg_config=rfg_config,
            chrom_sizes=chrom_sizes,
            pm=pm,
        )
    except BedBossException as e:
        output_bigbed = None
    if pm_clean:
        pm.stop_pipeline()

    _LOGGER.info(f"Bed output file: {output_bed}")
    _LOGGER.info(f"BigBed output file: {output_bigbed}")

    return BedMakerOutput(
        bed_file=output_bed,
        bigbed_file=os.path.abspath(output_bigbed) if output_bigbed else None,
        bed_digest=RegionSet(output_bed).identifier,
        bed_type=bed_type,
        bed_format=bed_format,
    )



import logging
import os
import tempfile
import pypiper
import subprocess

from bedboss.const import MAX_FILE_SIZE, MAX_REGION_NUMBER, MIN_REGION_WIDTH
from bedboss.exceptions import QualityException


_LOGGER = logging.getLogger("bedboss")


def bedqc(
    bedfile: str,
    outfolder: str,
    max_file_size: int = MAX_FILE_SIZE,
    max_region_number: int = MAX_REGION_NUMBER,
    min_region_width: int = MIN_REGION_WIDTH,
    pm: pypiper.PipelineManager = None,
) -> bool:
    """
    Perform quality checks on a BED file.

    :param bedfile: path to the bed file
    :param outfolder: path to the folder where to store information about pipeline and logs
    :param max_file_size: Maximum file size threshold to pass the quality check.
    :param max_region_number: Maximum number of regions threshold to pass the quality check.
    :param min_region_width: Minimum region width threshold to pass the quality check.
    :param pm: Pypiper object for managing pipeline operations.
    :return: True if the file passes the quality check.
    :raises QualityException: if the file does not pass the quality
    """
    _LOGGER.info("Running bedqc...")

    output_file = os.path.join(outfolder, "failed_qc.csv")
    bedfile_name = os.path.basename(bedfile)
    input_extension = os.path.splitext(bedfile_name)[1]

    # file_exists = os.path.isfile(bedfile)
    if not os.path.exists(outfolder):
        os.makedirs(outfolder)

    # to execute bedqc from inside Python (without using cli) Pypiper is set to default:
    if not pm:
        pm = pypiper.PipelineManager(
            name="bedQC-pipeline", outfolder=outfolder, recover=True, multi=True
        )

    detail = []

    # check number of regions
    script_path = os.path.join(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
        "bedqc",
        "est_line.sh",
    )
    assert os.path.exists(script_path), FileNotFoundError(
        f"'{script_path}' script not found"
    )

    if input_extension == ".gz":
        file = os.path.join(outfolder, next(tempfile._get_candidate_names()))
        pm.clean_add(file)
        cmd = "zcat " + bedfile + " > " + file
        pm.run(cmd, file)
    else:
        file = bedfile

    cmd = f"bash {script_path} {file} "

    if int(pm.checkprint(cmd)) > max_region_number:
        detail.append("File contains more than 5 million regions.")

    # check file size
    if os.path.getsize(bedfile) >= max_file_size:
        detail.append("File size is larger than 2G.")

    # check mean region width
    awk_command = r"""
    {
    diff += $3 - $2
    }
    END {
    print diff/NR
    }
    """

    if (
        float(subprocess.check_output(["awk", awk_command, file], text=True).split()[0])
        < min_region_width
    ):
        detail.append(f"Mean region width is less than {min_region_width} bp.")

    if len(detail) > 0:
        _LOGGER.info("file_name: ", bedfile_name)
        if os.path.exists(output_file):
            with open(output_file, "a") as f:
                f.write(f"{bedfile_name}\t{detail} \n")
        else:
            with open(output_file, "w") as f:
                f.write("file_name\tdetail \n")
                f.write(f"{bedfile_name}\t{detail} \n")

        raise QualityException(f"{str(detail)}")

    _LOGGER.info(f"File ({file}) has passed Quality Control!")
    return True



from pydantic import BaseModel, ConfigDict, Field
from typing import Dict

from enum import Enum
import pypiper
import pathlib
from typing import Union
from bbconf.models.bed_models import (
    BedStatsModel,
    BedPlots,
    BedFiles,
    BedClassification,
)

from bedboss.const import MAX_FILE_SIZE, MAX_REGION_NUMBER, MIN_REGION_WIDTH


class FILE_TYPE(str, Enum):
    BED = "bed"
    NARROWPEAK = "narrowpeak"
    BROADPEAK = "broadpeak"


class BedMetadata(BaseModel):
    sample_name: str
    genome: str
    organism: str = ""
    species_id: str = ""
    cell_type: str = ""
    cell_line: str = ""
    exp_protocol: str = Field("", description="Experimental protocol (e.g. ChIP-seq)")
    library_source: str = Field(
        "", description="Library source (e.g. genomic, transcriptomic)"
    )
    genotype: str = Field("", description="Genotype of the sample")
    target: str = Field("", description="Target of the assay (e.g. H3K4me3)")
    antibody: str = Field("", description="Antibody used in the assay")
    treatment: str = Field(
        "", description="Treatment of the sample (e.g. drug treatment)"
    )
    tissue: str = Field("", description="Tissue type")
    global_sample_id: str = Field("", description="Global sample identifier")
    global_experiment_id: str = Field("", description="Global experiment identifier")
    description: str = Field("", description="Description of the sample")

    model_config = ConfigDict(
        populate_by_name=True,
        extra="allow",
    )


class BedStatCLIModel(BaseModel):
    """
    CLI model for bedstat
    """

    bedfile: Union[str, pathlib.Path]
    genome: str
    outfolder: Union[str, pathlib.Path]
    bed_digest: str = None
    bigbed: Union[str, pathlib.Path] = None
    ensdb: str = None
    open_signal_matrix: str = None
    just_db_commit: bool = False
    pm: pypiper.PipelineManager = None

    model_config = ConfigDict(extra="ignore", arbitrary_types_allowed=True)


class BedQCCLIModel(BaseModel):
    """
    CLI model for bedqc
    """

    bedfile: Union[str, pathlib.Path]
    outfolder: Union[str, pathlib.Path]
    max_file_size: int = MAX_FILE_SIZE
    max_region_number: int = MAX_REGION_NUMBER
    min_region_width: int = MIN_REGION_WIDTH
    pm: pypiper.PipelineManager = None

    model_config = ConfigDict(extra="ignore", arbitrary_types_allowed=True)


class BedMakerCLIModel(BaseModel):
    """
    CLI model for bedmaker
    """

    input_file: Union[str, pathlib.Path]
    input_type: str
    output_path: Union[str, pathlib.Path]
    # output_bigbed: Union[str, pathlib.Path]
    # sample_name: str
    genome: str
    rfg_config: Union[str, pathlib.Path] = None
    chrom_sizes: str = None
    narrowpeak: bool = False
    # standardize: bool = False
    check_qc: bool = True
    pm: pypiper.PipelineManager = None

    model_config = ConfigDict(extra="ignore", arbitrary_types_allowed=True)


class StatsUpload(BedStatsModel):
    model_config = ConfigDict(extra="ignore")


class PlotsUpload(BedPlots):
    model_config = ConfigDict(extra="ignore")


class FilesUpload(BedFiles):
    model_config = ConfigDict(extra="ignore")


class BedClassificationUpload(BedClassification):
    model_config = ConfigDict(extra="ignore")



from pydantic import BaseModel, Field
from typing import Union
from pathlib import Path
from enum import Enum


class InputTypes(Enum):
    BED_GRAPH = "bedgraph"
    BIG_BED = "bigbed"
    BIG_WIG = "bigwig"
    WIG = "wig"
    BED = "bed"


class BedType(str, Enum):
    BED = "bed"
    NARROWPEAK = "narrowpeak"
    BROADPEAK = "broadpeak"


class BedMakerOutput(BaseModel):
    bed_file: Union[str, Path]
    bigbed_file: Union[str, Path, None] = None
    bed_digest: str = None
    bed_type: str = Field(
        default="bed3", pattern="^bed(?:[3-9]|1[0-5])(?:\+|$)[0-9]?+$"
    )
    bed_format: BedType = BedType.BED



""" Package-level data """

import logmuse
import coloredlogs
import logging


from bedboss._version import __version__


__package_name__ = "bedboss"

__author__ = [
    "Oleksandr Khoroshevskyi",
    "Michal Stolarczyk",
    "Ognen Duzlevski",
    "Jose Verdezoto",
    "Bingjie Xue",
    "Donald Campbell",
]
__email__ = "khorosh@virginia.edu"

__all__ = [
    "__version__",
    "__package_name__",
    "__author__",
]

_LOGGER = logmuse.init_logger("bedboss")
coloredlogs.install(
    logger=_LOGGER,
    datefmt="%H:%M:%S",
    fmt="[%(levelname)s] [%(asctime)s] [BEDBOSS] %(message)s",
)

_LOGGER_PIPESTAT = logging.getLogger("pipestat")
coloredlogs.install(
    logger=_LOGGER_PIPESTAT,
    datefmt="%H:%M:%S",
    fmt="[%(levelname)s] [%(asctime)s] [PIPESTAT] %(message)s",
)

_LOGGER_GENIML = logging.getLogger("geniml")
coloredlogs.install(
    logger=_LOGGER_GENIML,
    datefmt="%H:%M:%S",
    fmt="[%(levelname)s] [%(asctime)s] [GENIML] %(message)s",
)

_LOGGER_BBCONF = logging.getLogger("bbconf")
coloredlogs.install(
    logger=_LOGGER_BBCONF,
    datefmt="%H:%M:%S",
    fmt="[%(levelname)s] [%(asctime)s] [BBCONF] %(message)s",
)

_LOGGER_BBCONF = logging.getLogger("pephubclient")
coloredlogs.install(
    logger=_LOGGER_BBCONF,
    datefmt="%H:%M:%S",
    fmt="[%(levelname)s] [%(asctime)s] [PEPHUBCLIENT] %(message)s",
)



import logging
from typing import Optional, Tuple

import pandas.errors
import pandas as pd

from bedboss.exceptions import BedTypeException

_LOGGER = logging.getLogger("bedboss")


def get_bed_type(bed: str, no_fail: Optional[bool] = True) -> Tuple[str, str]:
    """
    get the bed file type (ex. bed3, bed3+n )
    standardize chromosomes if necessary:
    filter the input file to contain only the standard chromosomes,
    remove regions on ChrUn chromosomes

    :param bed: path to the bed file
    :param no_fail: should the function (and pipeline) continue if this function fails to parse BED file
    :return bedtype: tuple[option ["bed{bedtype}+{n}", "unknown_bedtype"], option [bed, narrowpeak, broadpeak, unknown_bedtype]]
    """
    #    column format for bed12
    #    string chrom;       "Reference sequence chromosome or scaffold"
    #    uint   chromStart;  "Start position in chromosome"
    #    uint   chromEnd;    "End position in chromosome"
    #    string name;        "Name of item."
    #    uint score;          "Score (0-1000)"
    #    char[1] strand;     "+ or - for strand"
    #    uint thickStart;   "Start of where display should be thick (start codon)"
    #    uint thickEnd;     "End of where display should be thick (stop codon)"
    #    uint reserved;     "Used as itemRgb as of 2004-11-22"
    #    int blockCount;    "Number of blocks"
    #    int[blockCount] blockSizes; "Comma separated list of block sizes"
    #    int[blockCount] chromStarts; "Start positions relative to chromStart"

    df = None

    max_rows = 5
    row_count = 0

    while row_count <= max_rows:
        try:
            df = pd.read_csv(bed, sep="\t", header=None, nrows=4, skiprows=row_count)
            if row_count > 0:
                _LOGGER.info(f"Skipped {row_count} rows to parse bed file {bed}")
            break
        except UnicodeDecodeError as e:
            try:
                df = pd.read_csv(
                    bed,
                    sep="\t",
                    header=None,
                    nrows=4,
                    skiprows=row_count,
                    encoding="utf-16",
                )
                if row_count > 0:
                    _LOGGER.info(f"Skipped {row_count} rows to parse bed file {bed}")
                break
            except (pandas.errors.ParserError, pandas.errors.EmptyDataError) as e:
                if row_count <= max_rows:
                    row_count += 1
                else:
                    if no_fail:
                        _LOGGER.warning(
                            f"Unable to parse bed file {bed}, due to error {e}, setting bed_type = unknown_bedtype"
                        )
                        return "unknown_bedtype", "unknown_bedtype"
                    else:
                        raise BedTypeException(
                            reason=f"Bed type could not be determined due to CSV parse error {e}"
                        )
        except (pandas.errors.ParserError, pandas.errors.EmptyDataError) as e:
            if row_count <= max_rows:
                row_count += 1
            else:
                if no_fail:
                    _LOGGER.warning(
                        f"Unable to parse bed file {bed}, due to error {e}, setting bed_type = unknown_bedtype"
                    )
                    return "unknown_bedtype", "unknown_bedtype"
                else:
                    raise BedTypeException(
                        reason=f"Bed type could not be determined due to CSV parse error {e}"
                    )

    if df is not None:
        df = df.dropna(axis=1)
        num_cols = len(df.columns)
        bedtype = 0

        if num_cols == 9 and ("broadpeak" in bed or "broadPeak" in bed):
            bed_type_named = "broadpeak"
        elif num_cols == 10 and ("narrowpeak" in bed or "narrowPeak" in bed):
            bed_type_named = "narrowpeak"
        else:
            bed_type_named = "bed"

        for col in df:
            if col <= 2:
                if col == 0:
                    if df[col].dtype == "O":
                        bedtype += 1
                    elif df[col].dtype == "int" or df[col].dtype == "float":
                        bedtype += 1
                    else:
                        if no_fail:
                            _LOGGER.warning(
                                f"Bed type could not be determined at column {0} with data type: {df[col].dtype}"
                            )
                            return "unknown_bedtype", "unknown_bedtype"
                        else:
                            raise BedTypeException(
                                reason=f"Bed type could not be determined at column {0} with data type: {df[col].dtype}"
                            )

                else:
                    if df[col].dtype == "int" and (df[col] >= 0).all():
                        bedtype += 1
                    else:
                        if no_fail:
                            _LOGGER.warning(
                                f"Bed type could not be determined at column {col} with data type: {df[col].dtype}"
                            )
                            return "unknown_bedtype", "unknown_bedtype"
                        else:
                            raise BedTypeException(
                                reason=f"Bed type could not be determined at column 0 with data type: {df[col].dtype}"
                            )
            else:
                if col == 3:
                    if df[col].dtype == "O":
                        bedtype += 1
                    else:
                        n = num_cols - bedtype
                        return f"bed{bedtype}+{n}", bed_type_named
                elif col == 4:
                    if df[col].dtype == "int" and df[col].between(0, 1000).all():
                        bedtype += 1
                    else:
                        n = num_cols - bedtype
                        return f"bed{bedtype}+{n}", bed_type_named
                elif col == 5:
                    if df[col].isin(["+", "-", "."]).all():
                        bedtype += 1
                    else:
                        n = num_cols - bedtype
                        return f"bed{bedtype}+{n}", bed_type_named
                elif 6 <= col <= 8:
                    if df[col].dtype == "int" and (df[col] >= 0).all():
                        # TODO Should we be increasing bedtype after 6?
                        bedtype += 1
                    elif num_cols == 10:
                        # This is a catch to see if this is actually a narrowpeak file that is unnamed
                        if col == 6 and all(
                            [
                                (df[col].dtype == "float" or df[col][0] == -1),
                                (df[col + 1].dtype == "float" or df[col + 1][0] == -1),
                                (df[col + 2].dtype == "float" or df[col + 2][0] == -1),
                                (df[col + 3].dtype == "int" or df[col + 3][0] == -1),
                            ]
                        ):
                            n = num_cols - bedtype
                            bed_type_named = "narrowpeak"
                            return f"bed{bedtype}+{n}", bed_type_named
                        else:
                            n = num_cols - bedtype
                            return f"bed{bedtype}+{n}", bed_type_named

                    elif num_cols == 9:
                        # This is a catch to see if this is actually a broadpeak file that is unnamed

                        if all(
                            [
                                (df[col].dtype == "float" or df[col][0] == -1),
                                (df[col + 1].dtype == "float" or df[col + 1][0] == -1),
                                (df[col + 2].dtype == "float" or df[col + 2][0] == -1),
                            ]
                        ):
                            n = num_cols - bedtype
                            bed_type_named = "broadpeak"
                            return f"bed{bedtype}+{n}", bed_type_named
                        else:
                            n = num_cols - bedtype
                            return f"bed{bedtype}+{n}", bed_type_named
                    else:
                        n = num_cols - bedtype
                        return f"bed{bedtype}+{n}", bed_type_named
                elif col == 9:
                    if df[col].dtype == "int":
                        bedtype += 1
                    else:
                        n = num_cols - bedtype
                        return f"bed{bedtype}+{n}", bed_type_named
                elif col == 10 or col == 11:
                    if df[col].str.match(r"^(\d+(,\d+)*)?$").all():
                        bedtype += 1
                    else:
                        n = num_cols - bedtype
                        return f"bed{bedtype}+{n}", bed_type_named
                else:
                    n = num_cols - bedtype
                    return f"bed{bedtype}+{n}", bed_type_named

        # This is to catch any files that are assigned a bed number but don't adhere to the above conditions
        return f"bed{bedtype}+0", bed_type_named

    else:
        return "unknown_bedtype", "unknown_bedtype"



import os
import pytest

from bedboss.bedclassifier import get_bed_type


FILE_DIR = os.path.dirname(os.path.realpath(__file__))
HG19_CORRECT_DIR = os.path.join(FILE_DIR, "test_data", "bed", "hg19", "correct")
FILE_PATH = f"{HG19_CORRECT_DIR}/sample1.bed.gz"
FILE_PATH_UNZIPPED = f"{HG19_CORRECT_DIR}/hg19_example1.bed"

SIMPLE_EXAMPLES_DIR = os.path.join(FILE_DIR, "data", "bed", "simpleexamples")
BED1 = f"{SIMPLE_EXAMPLES_DIR}/bed1.bed"
BED2 = f"{SIMPLE_EXAMPLES_DIR}/bed2.bed"
BED3 = f"{SIMPLE_EXAMPLES_DIR}/bed3.bed"


class TestBedClassifier:
    def test_classification(
        self,
    ):
        bedclass = get_bed_type(bed=FILE_PATH)

    def test_get_bed_type(
        self,
    ):
        bedtype = get_bed_type(bed=FILE_PATH_UNZIPPED)
        assert bedtype == ("bed6+3", "broadpeak")

    @pytest.mark.parametrize(
        "values",
        [
            (BED1, ("bed6+4", "narrowpeak")),
            (BED2, ("bed6+3", "broadpeak")),
            (BED3, ("bed6+2", "bed")),
        ],
    )
    def test_get_bed_types(self, values):
        # bed1 is encode narrowpeak
        # bed2 is encode broadpeak
        # bed 3 is encode bed6+ (6+2)

        bedclass = get_bed_type(bed=values[0])
        assert bedclass == values[1]

    @pytest.mark.skip(reason="Not implemented")
    def test_from_PEPhub_beds(
        self,
    ):
        """"""
        # TODO implement testing from pephub
        pass


#
# def test_manual_dir_beds():
#     """This test is currently just for local manual testing"""
#     local_dir = "/home/drc/Downloads/test_beds_BED_classifier/"
#     # local_dir = "/home/drc/Downloads/individual_beds/"
#     # local_dir = "/home/drc/Downloads/only_narrowpeaks/"
#     output_dir = "/home/drc/Downloads/BED_CLASSIFIER_OUTPUT/"
#     # local_dir = "/home/drc/Downloads/encode_beds/bedfiles/"
#     # output_dir = "/home/drc/Downloads/encode_beds/output/"
#     # local_dir = "/home/drc/Downloads/single_encode_beds/bedfiles/"
#     # output_dir = "/home/drc/Downloads/single_encode_beds/output/"
#
#     for root, dirs, files in os.walk(local_dir):
#         for file in files:
#             print(file)
#             file_path = os.path.join(root, file)
#             print(file_path)
#             bedclass = BedClassifier(
#                 input_file=file_path, output_dir=output_dir, bed_digest=file
#             )
#             print("\nDEBUG BEDCLASS\n")
#             print(bedclass.bed_type)
#             print("+++++++++++++++++++")
#
#
# #
# if __name__ == "__main__":
#     # test_get_bed_type()
#     # test_classification()
#     test_manual_dir_beds()



import typer
from typing import Union
import os

from bedboss import __version__
from bedboss.const import MAX_FILE_SIZE, MAX_REGION_NUMBER, MIN_REGION_WIDTH

# commented and made new const here, because it speeds up help function,
# from bbconf.const import DEFAULT_LICENSE
DEFAULT_LICENSE = "DUO:0000042"

app = typer.Typer(pretty_exceptions_short=False, pretty_exceptions_show_locals=False)


def create_pm(
    outfolder: str, multi: bool = False, recover: bool = True, dirty: bool = False
):
    import pypiper

    pm_out_folder = outfolder
    pm_out_folder = os.path.join(pm_out_folder, "pipeline_manager")

    pm: pypiper.PipelineManager = pypiper.PipelineManager(
        name="bedboss-pipeline",
        outfolder=pm_out_folder,
        version=__version__,
        multi=multi,
        recover=recover,
        dirty=dirty,
    )
    return pm


options_list = ["bigwig", "bedgraph", "bed", "bigbed", "wig"]


def validate_input_options(option: str):
    if option not in options_list:
        raise typer.BadParameter(
            f"Invalid input type option '{option}'. Options are: {', '.join(options_list)}"
        )
    return option


@app.command(help="Run all the bedboss pipeline for a single bed file")
def run_all(
    input_file: str = typer.Option(
        ...,
        help="Path to the input file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    input_type: str = typer.Option(
        ...,
        help=f"Type of the input file. Options are: {', '.join(options_list)}",
        callback=validate_input_options,
        case_sensitive=False,
    ),
    outfolder: str = typer.Option(..., help="Path to the output folder"),
    genome: str = typer.Option(..., help="Genome name. Example: 'hg38'"),
    bedbase_config: str = typer.Option(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    license_id: str = typer.Option(
        DEFAULT_LICENSE,
        help="License ID. If not provided for in PEP"
        "for each bed file, this license will be used",
    ),
    rfg_config: str = typer.Option(None, help="Path to the rfg config file"),
    narrowpeak: bool = typer.Option(False, help="Is the input file a narrowpeak file?"),
    check_qc: bool = typer.Option(True, help="Check the quality of the input file?"),
    chrom_sizes: str = typer.Option(None, help="Path to the chrom sizes file"),
    open_signal_matrix: str = typer.Option(
        None, help="Path to the open signal matrix file"
    ),
    ensdb: str = typer.Option(None, help="Path to the EnsDb database file"),
    just_db_commit: bool = typer.Option(False, help="Just commit to the database?"),
    force_overwrite: bool = typer.Option(
        False, help="Force overwrite the output files"
    ),
    upload_qdrant: bool = typer.Option(False, help="Upload to Qdrant"),
    upload_s3: bool = typer.Option(False, help="Upload to S3"),
    upload_pephub: bool = typer.Option(False, help="Upload to PEPHub"),
    # Universes
    universe: bool = typer.Option(False, help="Create a universe"),
    universe_method: str = typer.Option(
        None, help="Method used to create the universe"
    ),
    universe_bedset: str = typer.Option(
        None, help="Bedset used used to create the universe"
    ),
    # PipelineManager
    multi: bool = typer.Option(False, help="Run multiple samples"),
    recover: bool = typer.Option(True, help="Recover from previous run"),
    dirty: bool = typer.Option(False, help="Run without removing existing files"),
):
    """

    Run the bedboss pipeline for a single bed file
    """
    from bedboss.bedboss import run_all as run_all_bedboss
    from bbconf.bbagent import BedBaseAgent

    agent = BedBaseAgent(bedbase_config)

    run_all_bedboss(
        input_file=input_file,
        input_type=input_type,
        outfolder=outfolder,
        genome=genome,
        bedbase_config=agent,
        license_id=license_id,
        rfg_config=rfg_config,
        narrowpeak=narrowpeak,
        check_qc=check_qc,
        chrom_sizes=chrom_sizes,
        open_signal_matrix=open_signal_matrix,
        ensdb=ensdb,
        other_metadata=None,
        just_db_commit=just_db_commit,
        force_overwrite=force_overwrite,
        upload_qdrant=upload_qdrant,
        upload_s3=upload_s3,
        upload_pephub=upload_pephub,
        universe=universe,
        universe_method=universe_method,
        universe_bedset=universe_bedset,
        pm=create_pm(outfolder=outfolder, multi=multi, recover=recover, dirty=dirty),
    )


@app.command(help="Run the all bedboss pipeline for a bed files in a PEP")
def run_pep(
    pep: str = typer.Option(..., help="PEP file. Local or remote path"),
    outfolder: str = typer.Option(..., help="Path to the output folder"),
    bedbase_config: str = typer.Option(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    create_bedset: bool = typer.Option(False, help="Create a new bedset"),
    bedset_heavy: bool = typer.Option(
        False, help="Run the heavy version of the bedbuncher pipeline"
    ),
    bedset_id: Union[str, None] = typer.Option(None, help="Bedset ID"),
    rfg_config: str = typer.Option(None, help="Path to the rfg config file"),
    check_qc: bool = typer.Option(True, help="Check the quality of the input file?"),
    ensdb: str = typer.Option(None, help="Path to the EnsDb database file"),
    just_db_commit: bool = typer.Option(False, help="Just commit to the database?"),
    force_overwrite: bool = typer.Option(
        False, help="Force overwrite the output files"
    ),
    upload_qdrant: bool = typer.Option(False, help="Upload to Qdrant"),
    upload_s3: bool = typer.Option(False, help="Upload to S3"),
    upload_pephub: bool = typer.Option(False, help="Upload to PEPHub"),
    no_fail: bool = typer.Option(False, help="Do not fail on error"),
    license_id: str = typer.Option(DEFAULT_LICENSE, help="License ID"),
    # PipelineManager
    multi: bool = typer.Option(False, help="Run multiple samples"),
    recover: bool = typer.Option(True, help="Recover from previous run"),
    dirty: bool = typer.Option(False, help="Run without removing existing files"),
):
    """
    Run the bedboss pipeline for a bed files in a PEP
    """
    from bedboss.bedboss import insert_pep

    insert_pep(
        bedbase_config=bedbase_config,
        output_folder=outfolder,
        pep=pep,
        bedset_id=bedset_id,
        rfg_config=rfg_config,
        create_bedset=create_bedset,
        bedset_heavy=bedset_heavy,
        check_qc=check_qc,
        ensdb=ensdb,
        just_db_commit=just_db_commit,
        force_overwrite=force_overwrite,
        license_id=license_id,
        upload_s3=upload_s3,
        upload_pephub=upload_pephub,
        upload_qdrant=upload_qdrant,
        no_fail=no_fail,
        pm=create_pm(
            outfolder=outfolder,
            multi=multi,
            recover=recover,
            dirty=dirty,
        ),
    )


@app.command(help=f"Create a bed files form a [{', '.join(options_list)}] file")
def make_bed(
    input_file: str = typer.Option(
        ...,
        help="Path to the input file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    input_type: str = typer.Option(
        ...,
        help=f"Type of the input file. Options are: {', '.join(options_list)}",
        callback=validate_input_options,
        case_sensitive=False,
    ),
    outfolder: str = typer.Option(..., help="Path to the output folder"),
    genome: str = typer.Option(..., help="Genome name. Example: 'hg38'"),
    rfg_config: str = typer.Option(None, help="Path to the rfg config file"),
    narrowpeak: bool = typer.Option(False, help="Is the input file a narrowpeak file?"),
    chrom_sizes: str = typer.Option(None, help="Path to the chrom sizes file"),
    # PipelineManager
    multi: bool = typer.Option(False, help="Run multiple samples"),
    recover: bool = typer.Option(True, help="Recover from previous run"),
    dirty: bool = typer.Option(False, help="Run without removing existing files"),
):
    """

    Run the bedboss pipeline for a single bed file
    """
    from bedboss.bedmaker.bedmaker import make_bed as mk_bed_func

    mk_bed_func(
        input_file=input_file,
        input_type=input_type,
        output_path=outfolder,
        genome=genome,
        narrowpeak=narrowpeak,
        rfg_config=rfg_config,
        chrom_sizes=chrom_sizes,
        pm=create_pm(outfolder=outfolder, multi=multi, recover=recover, dirty=dirty),
    )


@app.command(help=f"Create a bigbed files form a bed file")
def make_bigbed(
    bed_file: str = typer.Option(
        ...,
        help="Path to the input file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    bed_type: str = typer.Option(
        ...,
        help="bed type to be used for bigBed file generation 'bed{bedtype}+{n}' [Default: None] (e.g bed3+1)",
    ),
    outfolder: str = typer.Option(..., help="Path to the output folder"),
    genome: str = typer.Option(..., help="Genome name. Example: 'hg38'"),
    rfg_config: str = typer.Option(None, help="Path to the rfg config file"),
    chrom_sizes: str = typer.Option(None, help="Path to the chrom sizes file"),
    # PipelineManager
    multi: bool = typer.Option(False, help="Run multiple samples"),
    recover: bool = typer.Option(True, help="Recover from previous run"),
    dirty: bool = typer.Option(False, help="Run without removing existing files"),
):
    """

    Run the bedboss pipeline for a single bed file
    """
    from bedboss.bedmaker.bedmaker import make_bigbed as mk_bigbed_func

    mk_bigbed_func(
        bed_path=bed_file,
        output_path=outfolder,
        genome=genome,
        bed_type=bed_type,
        rfg_config=rfg_config,
        chrom_sizes=chrom_sizes,
        pm=create_pm(outfolder=outfolder, multi=multi, recover=recover, dirty=dirty),
    )


@app.command(help="Run the quality control for a bed file")
def run_qc(
    bed_file: str = typer.Option(
        ...,
        help="Path to the bed file to check the quality control on.",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    outfolder: str = typer.Option(..., help="Path to the output folder"),
    max_file_size: int = typer.Option(
        MAX_FILE_SIZE, help="Maximum file size threshold to pass the quality"
    ),
    max_region_number: int = typer.Option(
        MAX_REGION_NUMBER,
        help="Maximum number of regions threshold to pass the quality",
    ),
    min_region_width: int = typer.Option(
        MIN_REGION_WIDTH, help="Minimum region width threshold to pass the quality"
    ),
    # PipelineManager
    multi: bool = typer.Option(False, help="Run multiple samples"),
    recover: bool = typer.Option(True, help="Recover from previous run"),
    dirty: bool = typer.Option(False, help="Run without removing existing files"),
):
    from bedboss.bedqc.bedqc import bedqc

    bedqc(
        bedfile=bed_file,
        outfolder=outfolder,
        max_file_size=max_file_size,
        max_region_number=max_region_number,
        min_region_width=min_region_width,
        pm=create_pm(outfolder=outfolder, multi=multi, recover=recover, dirty=dirty),
    )


@app.command(help="Create the statistics for a single bed file.")
def run_stats(
    bed_file: str = typer.Option(
        ...,
        help="Path to the bed file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    genome: str = typer.Option(..., help="Genome name. Example: 'hg38'"),
    outfolder: str = typer.Option(..., help="Path to the output folder"),
    ensdb: str = typer.Option(None, help="Path to the EnsDb database file"),
    open_signal_matrix: str = typer.Option(
        None, help="Path to the open signal matrix file"
    ),
    just_db_commit: bool = typer.Option(False, help="Just commit to the database?"),
    # PipelineManager
    multi: bool = typer.Option(False, help="Run multiple samples"),
    recover: bool = typer.Option(True, help="Recover from previous run"),
    dirty: bool = typer.Option(False, help="Run without removing existing files"),
):
    from bedboss.bedstat.bedstat import bedstat

    bedstat(
        bedfile=bed_file,
        genome=genome,
        outfolder=outfolder,
        ensdb=ensdb,
        open_signal_matrix=open_signal_matrix,
        just_db_commit=just_db_commit,
        pm=create_pm(outfolder=outfolder, multi=multi, recover=recover, dirty=dirty),
    )


@app.command(
    help="Reindex the bedbase database and insert all files to the qdrant database."
)
def reindex(
    bedbase_config: str = typer.Option(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
):
    from bedboss.qdrant_index.qdrant_index import add_to_qdrant

    add_to_qdrant(config=bedbase_config)


@app.command(
    help="Create a bedset from a pep file, and insert it to the bedbase database."
)
def make_bedset(
    pep: str = typer.Option(..., help="PEP file. Local or remote path"),
    outfolder: str = typer.Option(..., help="Path to the output folder"),
    bedbase_config: str = typer.Option(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    bedset_name: str = typer.Option(..., help="Name of the bedset"),
    heavy: bool = typer.Option(False, help="Run the heavy version of the pipeline"),
    force_overwrite: bool = typer.Option(
        False, help="Force overwrite the output files"
    ),
    upload_s3: bool = typer.Option(False, help="Upload to S3"),
    upload_pephub: bool = typer.Option(False, help="Upload to PEPHub"),
    no_fail: bool = typer.Option(False, help="Do not fail on error"),
):
    from bedboss.bedbuncher.bedbuncher import run_bedbuncher_form_pep

    run_bedbuncher_form_pep(
        bedbase_config=bedbase_config,
        bedset_pep=pep,
        output_folder=outfolder,
        bedset_name=bedset_name,
        heavy=heavy,
        upload_pephub=upload_pephub,
        upload_s3=upload_s3,
        no_fail=no_fail,
        force_overwrite=force_overwrite,
    )


@app.command(help="Initialize the new, sample configuration file")
def init_config(
    outfolder: str = typer.Option(..., help="Path to the output folder"),
):
    from bedboss.utils import save_example_bedbase_config

    save_example_bedbase_config(outfolder)


@app.command(help="Delete bed from the bedbase database")
def delete_bed(
    sample_id: str = typer.Option(..., help="Sample ID"),
    config: str = typer.Option(..., help="Path to the bedbase config file"),
):
    from bbconf import BedBaseAgent

    bbagent = BedBaseAgent(config)
    bbagent.bed.delete(sample_id)
    print(f"sample {sample_id} deleted from the bedbase database")


def version_callback(value: bool):
    if value:
        typer.echo(f"Bedboss version: {__version__}")
        raise typer.Exit()


@app.command(help="Delete BedSet from the bedbase database")
def delete_bedset(
    identifier: str = typer.Option(..., help="BedSet ID"),
    config: str = typer.Option(..., help="Path to the bedbase config file"),
):
    from bbconf import BedBaseAgent

    bbagent = BedBaseAgent(config)
    bbagent.bedset.delete(identifier)
    print(f"BedSet {identifier} deleted from the bedbase database")


@app.command(help="Tokenize a bedfile")
def tokenize_bed(
    bed_id: str = typer.Option(
        ...,
        help="Path to the bed file",
    ),
    universe_id: str = typer.Option(
        ...,
        help="Universe ID",
    ),
    cache_folder: str = typer.Option(
        None,
        help="Path to the cache folder",
    ),
    add_to_db: bool = typer.Option(
        False,
        help="Add the tokenized bed file to the bedbase database",
    ),
    bedbase_config: str = typer.Option(
        None,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    overwrite: bool = typer.Option(
        False,
        help="Overwrite the existing tokenized bed file",
    ),
):
    from bedboss.tokens.tokens import tokenize_bed_file

    tokenize_bed_file(
        universe=universe_id,
        bed=bed_id,
        cache_folder=cache_folder,
        add_to_db=add_to_db,
        config=bedbase_config,
        overwrite=overwrite,
    )


@app.command(help="Delete tokenized bed file")
def delete_tokenized(
    universe_id: str = typer.Option(
        ...,
        help="Universe ID",
    ),
    bed_id: str = typer.Option(
        ...,
        help="Bed ID",
    ),
    config: str = typer.Option(
        None,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
):
    from bedboss.tokens.tokens import delete_tokenized

    delete_tokenized(
        universe=universe_id,
        bed=bed_id,
        config=config,
    )


@app.command(help="Convert bed file to universe")
def convert_universe(
    bed_id: str = typer.Option(
        ...,
        help="Path to the bed file",
    ),
    config: str = typer.Option(
        ...,
        help="Path to the bedbase config file",
        exists=True,
        file_okay=True,
        readable=True,
    ),
    method: str = typer.Option(
        None,
        help="Method used to create the universe",
    ),
    bedset: str = typer.Option(
        None,
        help="Bedset used to create the universe",
    ),
):
    from bbconf.bbagent import BedBaseAgent

    bbagent = BedBaseAgent(config)
    bbagent.bed.add_universe(
        bedfile_id=bed_id,
        bedset_id=bedset,
        construct_method=method,
    )


@app.command(help="Check installed R packages")
def check_requirements():
    from bedboss.bedboss import requirements_check

    print("Checking piplines requirements...")
    requirements_check()


@app.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(
        None, "--version", "-v", callback=version_callback, help="App version"
    ),
):
    pass



import logging
import os
from typing import Union

import pypiper
import peppy
from eido import validate_project
import bbconf
import subprocess

import pephubclient
from pephubclient.helpers import is_registry_path, MessageHandler as m
from bbconf.bbagent import BedBaseAgent
from bbconf.models.base_models import FileModel
from bbconf.const import DEFAULT_LICENSE

from bedboss.bedstat.bedstat import bedstat
from bedboss.bedmaker.bedmaker import make_all
from bedboss.bedbuncher import run_bedbuncher
from bedboss.const import (
    BEDBOSS_PEP_SCHEMA_PATH,
    PKG_NAME,
)
from bedboss.models import (
    StatsUpload,
    PlotsUpload,
    FilesUpload,
    BedClassificationUpload,
)
from bedboss.utils import (
    standardize_genome_name,
    get_genome_digest,
)
from bedboss.exceptions import BedBossException
from bedboss._version import __version__

_LOGGER = logging.getLogger(PKG_NAME)


def requirements_check() -> None:
    """
    Check if all requirements are installed

    :return: None
    """
    _LOGGER.info("Checking requirements...")
    subprocess.run(
        ["bash", f"{os.path.dirname(os.path.abspath(__file__))}/requirements_test.sh"]
    )


def run_all(
    input_file: str,
    input_type: str,
    outfolder: str,
    genome: str,
    bedbase_config: Union[str, bbconf.BedBaseAgent],
    name: str = None,
    license_id: str = DEFAULT_LICENSE,
    rfg_config: str = None,
    narrowpeak: bool = False,
    check_qc: bool = True,
    chrom_sizes: str = None,
    open_signal_matrix: str = None,
    ensdb: str = None,
    other_metadata: dict = None,
    just_db_commit: bool = False,
    force_overwrite: bool = False,
    upload_qdrant: bool = False,
    upload_s3: bool = False,
    upload_pephub: bool = False,
    # Universes
    universe: bool = False,
    universe_method: str = None,
    universe_bedset: str = None,
    pm: pypiper.PipelineManager = None,
) -> str:
    """
    Run bedboss: bedmaker -> bedqc -> bedclassifier -> bedstat -> upload to s3, qdrant, pephub, and bedbase.

    :param str input_file: Input file [required]
    :param str input_type: Input type [required] options: (bigwig|bedgraph|bed|bigbed|wig)
    :param str outfolder: Folder, where output should be saved  [required]
    :param str genome: genome_assembly of the sample. [required] options: (hg19, hg38, mm10) # TODO: add more
    :param str name: name of the sample (human-readable name, e.g. "H3K27ac in liver") [optional]
    :param Union[str, bbconf.BedBaseConf] bedbase_config: The path to the bedbase configuration file, or bbconf object.
    :param str license_id: license identifier [optional] (default: "DUO:0000042").; Find All licenses in bedbase.org
    :param str rfg_config: file path to the genome config file [optional]
    :param bool narrowpeak: whether the regions are narrow. Used to create bed file from bedgraph or bigwig
        (transcription factor implies narrow, histone mark implies broad peaks) [optional]
    :param bool check_qc: set True to run quality control during badmaking [optional] (default: True)
    :param str chrom_sizes: a full path to the chrom.sizes required for the bedtobigbed conversion [optional]
    :param str open_signal_matrix: a full path to the openSignalMatrix required for the tissue [optional]
    :param dict other_metadata: a dict containing all attributes from the sample
    :param str ensdb: a full path to the ensdb gtf file required for genomes not in GDdata [optional]
        (basically genomes that's not in GDdata)
    :param bool just_db_commit: whether just to commit the JSON to the database (default: False)
    :param bool force_overwrite: force overwrite analysis (default: False)
    :param bool upload_qdrant: whether to skip qdrant indexing
    :param bool upload_s3: whether to upload to s3
    :param bool upload_pephub: whether to push bedfiles and metadata to pephub (default: False)

    :param bool universe: whether to add the sample as the universe [Default: False]
    :param str universe_method: method used to create the universe [Default: None]
    :param str universe_bedset: bedset identifier for the universe [Default: None]
    :param pypiper.PipelineManager pm: pypiper object
    :return str bed_digest: bed digest
    """
    if isinstance(bedbase_config, str):
        bbagent = BedBaseAgent(bedbase_config)
    elif isinstance(bedbase_config, bbconf.BedBaseAgent):
        bbagent = bedbase_config
    else:
        raise BedBossException("Incorrect bedbase_config type. Exiting...")

    genome = standardize_genome_name(genome)

    _LOGGER.info(f"Input file = '{input_file}'")
    _LOGGER.info(f"Output folder = '{outfolder}'")

    if not pm:
        pm_out_folder = os.path.join(os.path.abspath(outfolder), "pipeline_manager")
        _LOGGER.info(f"Pipeline info folder = '{pm_out_folder}'")
        pm = pypiper.PipelineManager(
            name="bedboss-pipeline",
            outfolder=pm_out_folder,
            version=__version__,
            recover=True,
        )
        stop_pipeline = True
    else:
        stop_pipeline = False

    bed_metadata = make_all(
        input_file=input_file,
        input_type=input_type,
        output_path=outfolder,
        genome=genome,
        rfg_config=rfg_config,
        narrowpeak=narrowpeak,
        check_qc=check_qc,
        chrom_sizes=chrom_sizes,
        pm=pm,
    )
    if not other_metadata:
        other_metadata = {}

    statistics_dict = bedstat(
        bedfile=bed_metadata.bed_file,
        outfolder=outfolder,
        genome=genome,
        ensdb=ensdb,
        bed_digest=bed_metadata.bed_digest,
        open_signal_matrix=open_signal_matrix,
        just_db_commit=just_db_commit,
        pm=pm,
    )
    statistics_dict["bed_type"] = bed_metadata.bed_type
    statistics_dict["bed_format"] = bed_metadata.bed_format.value

    if bed_metadata.bigbed_file:
        genome_digest = get_genome_digest(genome)
    else:
        genome_digest = None

    stats = StatsUpload(**statistics_dict)
    plots = PlotsUpload(**statistics_dict)

    if bed_metadata.bigbed_file:
        big_bed = FileModel(
            name="bigbedfile",
            title="BigBed file",
            path=bed_metadata.bigbed_file,
            description="Path to the bigbed file",
        )
    else:
        big_bed = None
    files = FilesUpload(
        bed_file=FileModel(
            name="bedfile",
            title="BED file",
            path=bed_metadata.bed_file,
            description="Path to the BED file",
        ),
        bigbed_file=big_bed,
    )

    classification = BedClassificationUpload(
        name=name or bed_metadata.bed_digest,
        genome_digest=genome_digest,
        genome_alias=genome,
        bed_type=bed_metadata.bed_type,
        bed_format=bed_metadata.bed_format.value,
    )

    bbagent.bed.add(
        identifier=bed_metadata.bed_digest,
        stats=stats.model_dump(exclude_unset=True),
        metadata=other_metadata,
        plots=plots.model_dump(exclude_unset=True),
        files=files.model_dump(exclude_unset=True),
        classification=classification.model_dump(exclude_unset=True),
        license_id=license_id,
        upload_qdrant=upload_qdrant,
        upload_pephub=upload_pephub,
        upload_s3=upload_s3,
        local_path=outfolder,
        overwrite=force_overwrite,
        nofail=True,
    )

    if universe:
        bbagent.bed.add_universe(
            bedfile_id=bed_metadata.bed_digest,
            bedset_id=universe_bedset,
            construct_method=universe_method,
        )

    if stop_pipeline:
        pm.stop_pipeline()

    _LOGGER.info(f"All done! Bed digest: {bed_metadata.bed_digest}")
    return bed_metadata.bed_digest


def insert_pep(
    bedbase_config: str,
    output_folder: str,
    pep: Union[str, peppy.Project],
    bedset_id: str = None,
    bedset_name: str = None,
    rfg_config: str = None,
    license_id: str = DEFAULT_LICENSE,
    create_bedset: bool = False,
    bedset_heavy: bool = False,
    check_qc: bool = True,
    ensdb: str = None,
    just_db_commit: bool = False,
    force_overwrite: bool = False,
    upload_s3: bool = False,
    upload_pephub: bool = False,
    upload_qdrant: bool = False,
    no_fail: bool = False,
    pm: pypiper.PipelineManager = None,
) -> None:
    """
    Run all bedboss pipelines for all samples in the pep file.
    bedmaker -> bedqc -> bedstat -> qdrant_indexing -> bedbuncher

    :param str bedbase_config: bedbase configuration file path
    :param str output_folder: output statistics folder
    :param Union[str, peppy.Project] pep: path to the pep file or pephub registry path
    :param str bedset_id: bedset identifier
    :param str bedset_name: bedset name
    :param str rfg_config: path to the genome config file (refgenie)
    :param str license_id: license identifier [optional] (default: "DUO:0000042").; Find All licenses in bedbase.org
        This license will be used for bedfiles where license is not provided in PEP file
    :param bool create_bedset: whether to create bedset
    :param bool bedset_heavy: whether to use heavy processing (add all columns to the database)
    :param bool upload_qdrant: whether to upload bedfiles to qdrant
    :param bool check_qc: whether to run quality control during badmaking
    :param str ensdb: a full path to the ensdb gtf file required for genomes not in GDdata
    :param bool just_db_commit: whether save only to the database (Without saving locally )
    :param bool force_overwrite: whether to overwrite the existing record
    :param bool upload_s3: whether to upload to s3
    :param bool upload_pephub: whether to push bedfiles and metadata to pephub (default: False)
    :param bool upload_qdrant: whether to execute qdrant indexing
    :param bool no_fail: whether to raise an error if bedset was not added to the database
    :param pypiper.PipelineManager pm: pypiper object
    :return: None
    """

    failed_samples = []
    processed_ids = []
    if isinstance(pep, peppy.Project):
        pass
    elif isinstance(pep, str):
        if is_registry_path(pep):
            pep = pephubclient.PEPHubClient().load_project(pep)
        else:
            pep = peppy.Project(pep)
    else:
        raise BedBossException("Incorrect pep type. Exiting...")

    bbagent = BedBaseAgent(bedbase_config)

    validate_project(pep, BEDBOSS_PEP_SCHEMA_PATH)

    for i, pep_sample in enumerate(pep.samples):
        m.print_success(f"Processing sample {i + 1}/{len(pep.samples)}")
        _LOGGER.info(f"Running bedboss pipeline for {pep_sample.sample_name}")
        if pep_sample.get("file_type"):
            if pep_sample.get("file_type").lower() == "narrowpeak":
                is_narrow_peak = True
            else:
                is_narrow_peak = False
        else:
            is_narrow_peak = False
        try:
            bed_id = run_all(
                input_file=pep_sample.input_file,
                input_type=pep_sample.input_type,
                genome=pep_sample.genome,
                name=pep_sample.sample_name,
                bedbase_config=bbagent,
                license_id=pep_sample.get("license_id") or license_id,
                narrowpeak=is_narrow_peak,
                chrom_sizes=pep_sample.get("chrom_sizes"),
                open_signal_matrix=pep_sample.get("open_signal_matrix"),
                other_metadata=pep_sample.to_dict(),
                outfolder=output_folder,
                rfg_config=rfg_config,
                check_qc=check_qc,
                ensdb=ensdb,
                just_db_commit=just_db_commit,
                force_overwrite=force_overwrite,
                upload_qdrant=upload_qdrant,
                upload_s3=upload_s3,
                upload_pephub=upload_pephub,
                universe=pep_sample.get("universe"),
                universe_method=pep_sample.get("universe_method"),
                universe_bedset=pep_sample.get("universe_bedset"),
                pm=pm,
            )

            processed_ids.append(bed_id)
        except BedBossException as e:
            _LOGGER.error(f"Failed to process {pep_sample.sample_name}. See {e}")
            failed_samples.append(pep_sample.sample_name)

    if create_bedset:
        _LOGGER.info(f"Creating bedset from {pep.name}")
        run_bedbuncher(
            bedbase_config=bbagent,
            record_id=bedset_id or pep.name,
            bed_set=processed_ids,
            name=bedset_name or pep.name,
            output_folder=output_folder,
            description=pep.description,
            heavy=bedset_heavy,
            upload_pephub=upload_pephub,
            upload_s3=upload_s3,
            no_fail=no_fail,
            force_overwrite=force_overwrite,
        )
    else:
        _LOGGER.info(
            f"Skipping bedset creation. Create_bedset is set to {create_bedset}"
        )
    m.print_success(f"Processed samples: {processed_ids}")
    m.print_error(f"Failed samples: {failed_samples}")

    m.print_success(f"Processed samples: {processed_ids}")
    m.print_error(f"Failed samples: {failed_samples}")

    return None



import os
import logging
import urllib.request
import requests
from pephubclient.files_manager import FilesManager


_LOGGER = logging.getLogger("bedboss")


def standardize_genome_name(input_genome: str) -> str:
    """
    Standardizing user provided genome

    :param input_genome: standardize user provided genome, so bedboss know what genome
    we should use
    :return: genome name string
    """
    input_genome = input_genome.strip().lower()
    # TODO: we have to add more genome options and preprocessing of the string
    if input_genome == "hg38" or input_genome == "grch38":
        return "hg38"
    elif input_genome == "hg19" or input_genome == "grch37":
        return "hg19"
    elif input_genome == "mm10":
        return "mm10"
    # else:
    #     raise GenomeException("Incorrect genome assembly was provided")
    else:
        return input_genome


def download_file(url: str, path: str, no_fail: bool = False) -> None:
    """
    Download file from the url to specific location

    :param url: URL of the file
    :param path: Local path with filename
    :param no_fail: If True, do not raise exception if download fails
    :return: NoReturn
    """
    _LOGGER.info(f"Downloading remote file: {url}")
    _LOGGER.info(f"Local path: {os.path.abspath(path)}")
    try:
        urllib.request.urlretrieve(url, path)
        _LOGGER.info("File downloaded successfully!")
    except Exception as e:
        _LOGGER.error("File download failed.")
        if not no_fail:
            raise e
        _LOGGER.error("File download failed. Continuing anyway...")


def get_genome_digest(genome: str) -> str:
    return requests.get(
        f"http://refgenomes.databio.org/genomes/genome_digest/{genome}"
    ).text.strip('""')


def example_bedbase_config():
    """
    Return example configuration for BedBase
    """
    return {
        "database": {
            "host": "localhost",
            "port": 5432,
            "user": "postgres",
            "password": "docker",
            "database": "bedbase",
            "dialect": "postgresql",
            "driver": "psycopg",
        },
        "qdrant": {
            "host": "localhost",
            "port": 6333,
            "api_key": None,
            "collection": "test_collection",
        },
        "server": {"host": "0.0.0.0", "port": 8000},
        "path": {
            "region2vec": "databio/r2v-encode-hg38",
            "vec2vec": "databio/v2v-geo-hg38",
            "text2vec": "sentence-transformers/all-MiniLM-L6-v2",
        },
        "access_methods": {
            "http": {
                "type": "https",
                "description": "HTTP compatible path",
                "prefix": "https://data2.bedbase.org/",
            },
            "s3": {
                "type": "s3",
                "description": "S3 compatible path",
                "prefix": "s3://data2.bedbase.org/",
            },
            "local": {
                "type": "https",
                "description": "How to serve local files.",
                "prefix": "/static/",
            },
        },
        "s3": {
            "endpoint_url": None,
            "aws_access_key_id": None,
            "aws_secret_access_key": None,
            "bucket": "bedbase",
        },
        "phc": {"namespace": "bedbase", "name": "bedbase", "tag": "latest"},
    }


def save_example_bedbase_config(path: str) -> None:
    """
    Save example configuration for BedBase

    :param path: path to the file
    """
    file_path = os.path.abspath(os.path.join(path, "bedbase_config.yaml"))
    FilesManager.save_yaml(example_bedbase_config(), file_path)
    _LOGGER.info(f"Example BedBase configuration saved to: {file_path}")
